/*
 * Copyright (c) 2009, 2023 Oracle and/or its affiliates. All rights reserved.
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License v. 2.0, which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * This Source Code may also be made available under the following Secondary
 * Licenses when the conditions for such availability set forth in the
 * Eclipse Public License v. 2.0 are satisfied: GNU General Public License,
 * version 2 with the GNU Classpath Exception, which is available at
 * https://www.gnu.org/software/classpath/license.html.
 *
 * SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
 */

package org.xbib.net.mail.test.test;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.PushbackInputStream;
import java.nio.charset.StandardCharsets;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * Handle protocol connection.
 *
 * Inspired by, and derived from, POP3Handler by sbo.
 *
 * @author sbo
 * @author Bill Shannon
 */
public abstract class ProtocolHandler {

    /**
     * Logger for this class.
     */
    private static final Logger logger = Logger.getLogger(ProtocolHandler.class.getName());

    /**
     * Quit?
     */
    protected boolean quit;

    public void handle(InputStream inputStream, OutputStream outputStream) {
        try {
            sendGreetings(outputStream);
            while (!quit) {
                handleCommand(inputStream, outputStream);
            }
            if (quit) {
                try {
                    if (inputStream != null)
                        inputStream.close();
                    if (outputStream != null) {
                        outputStream.close();
                    }
                } catch (final IOException e) {
                    logger.log(Level.SEVERE, "Error", e);
                }
            }
        } catch (IOException sex) {
            // ignore it, often get "connection reset" when client closes
        } catch (Exception e) {
            logger.log(Level.SEVERE, "Error", e);
        } finally {
            try {
                if (inputStream != null)
                    inputStream.close();
                if (outputStream != null) {
                    outputStream.close();
                }
            } catch (final IOException ioe) {
                logger.log(Level.SEVERE, "Error", ioe);
            }
        }
    }

    /**
     * Optionally send a greeting when first connected.
     */
    public void sendGreetings(OutputStream outputStream) throws IOException {
    }

    /**
     * Read and process a single command.
     */
    public abstract void handleCommand(InputStream inputStream, OutputStream outputStream) throws IOException;

    /**
     * Read a single line terminated by newline or CRLF.
     * Convert the UTF-8 bytes in the line (minus the line terminator)
     * to a String.
     */
    protected String readLine(InputStream inputStream) throws IOException {
        byte[] buf = new byte[128];
        int room = buf.length;
        int offset = 0;
        int c;
        while ((c = inputStream.read()) != -1) {
            if (c == '\n') {
                break;
            } else if (c == '\r') {
                int c2 = inputStream.read();
                if ((c2 != '\n') && (c2 != -1)) {
                    if (!(inputStream instanceof PushbackInputStream)) {
                        inputStream = new PushbackInputStream(inputStream);
                    }
                    ((PushbackInputStream) inputStream).unread(c2);
                }
                break;
            } else {
                if (--room < 0) {
                    byte[] nbuf = new byte[offset + 128];
                    room = nbuf.length - offset - 1;
                    System.arraycopy(buf, 0, nbuf, 0, offset);
                    buf = nbuf;
                }
                buf[offset++] = (byte) c;
            }
        }
        if ((c == -1) && (offset == 0)) {
            return null;
        }
        return new String(buf, 0, offset, StandardCharsets.UTF_8);
    }

    public void exit() {
        quit = true;
    }
}
